"""
various cucu utilities can be placed here and then exposed publicly through
the src/cucu/__init__.py
"""

import hashlib
import logging
import os
import pkgutil
import shutil
import time

import humanize
from selenium.webdriver.common.by import By
from tabulate import DataRow, TableFormat, tabulate
from tenacity import (
    after_log,
    before_log,
    before_sleep_log,
    retry_if_not_exception_type,
    stop_after_delay,
    wait_fixed,
)
from tenacity import retry as retrying

from cucu import logger
from cucu.browser.core import Browser
from cucu.config import CONFIG

GHERKIN_TABLEFORMAT = TableFormat(
    lineabove=None,
    linebelowheader=None,
    linebetweenrows=None,
    linebelow=None,
    headerrow=DataRow("|", "|", "|"),
    datarow=DataRow("|", "|", "|"),
    padding=1,
    with_header_hide=["lineabove"],
)


class StopRetryException(Exception):
    pass


def generate_short_id(seed=None, length=7):
    """
    Generate a short character ID based on seed (defaults to current performance counter).
    """
    if seed is None:
        seed = time.perf_counter()

    return hashlib.sha256(str(seed).encode("utf-8")).hexdigest()[:length]


def format_gherkin_table(table, headings=[], prefix=""):
    formatted = tabulate(table, headings, tablefmt=GHERKIN_TABLEFORMAT)
    if prefix == "":
        return formatted

    return prefix + formatted.replace("\n", f"\n{prefix}")


#
# code below adapted from:
# https://github.com/behave/behave/blob/994dbfe30e2a372182ea613333e06f069ab97d4b/behave/runner.py#L385
# so we can have the sub steps printed in the console logs
#
def run_steps(ctx, steps_text):
    """
    run sub steps within an existing step definition but also log their output
    so that its easy to see what is happening.
    """

    # -- PREPARE: Save original ctx data for current step.
    # Needed if step definition that called this method uses .table/.text
    original_table = getattr(ctx, "table", None)
    original_text = getattr(ctx, "text", None)

    # first time a given step calls substeps we want to move the step_index
    # so it starts the followings steps the right point.
    if not ctx.current_step.has_substeps:
        ctx.step_index += 1

    ctx.current_step.has_substeps = True

    ctx.feature.parser.variant = "steps"
    steps = ctx.feature.parser.parse_steps(steps_text)

    current_step = ctx.current_step
    current_step_start_at = current_step.start_at

    # XXX: I want to get back to this and find a slightly better way to handle
    #      these substeps without mucking around with so much state in behave
    #      but for now this works correctly and existing tests work as expected.
    try:
        with ctx._use_with_behave_mode():
            for step in steps:
                for formatter in ctx._runner.formatters:
                    step.is_substep = True
                    formatter.insert_step(step, index=ctx.step_index)

                passed = step.run(ctx._runner, quiet=False, capture=False)

                if not passed:
                    if "StopRetryException" in step.error_message:
                        raise StopRetryException(step.error_message)
                    else:
                        raise RuntimeError(step.error_message)

            # -- FINALLY: Restore original ctx data for current step.
            ctx.table = original_table
            ctx.text = original_text
    finally:
        ctx.current_step = current_step
        ctx.current_step.start_at = current_step_start_at

    return True


def retry(func, wait_up_to_s=None, retry_after_s=None):
    """
    utility retry function that can retry the provided `func` for the maximum
    amount of seconds specified by `wait_up_to_s` and wait the number of seconds
    specified in `retry_after_s`
    """
    if wait_up_to_s is None:
        wait_up_to_s = float(CONFIG["CUCU_STEP_WAIT_TIMEOUT_S"])

    if retry_after_s is None:
        retry_after_s = float(CONFIG["CUCU_STEP_RETRY_AFTER_S"])

    @retrying(
        stop=stop_after_delay(wait_up_to_s),
        wait=wait_fixed(retry_after_s),
        retry=retry_if_not_exception_type(StopRetryException),
        before=before_log(logger, logging.DEBUG),
        after=after_log(logger, logging.DEBUG),
        before_sleep=before_sleep_log(logger, logging.DEBUG),
    )
    def new_decorator(*args, **kwargs):
        ctx = CONFIG["__CUCU_CTX"]

        for hook in CONFIG["__CUCU_BEFORE_RETRY_HOOKS"]:
            hook(ctx)

        return func(*args, **kwargs)

    return new_decorator


def load_jquery_lib():
    """
    load jquery library
    """
    jquery_lib = pkgutil.get_data(
        "cucu", "external/jquery/jquery-3.5.1.min.js"
    )
    return jquery_lib.decode("utf8")


def text_in_current_frame(browser: Browser) -> str:
    """
    Utility to get all the visible text of the current frame.

    Args:
        browser (Browser): the browser session switched to the desired frame
    """
    script = "return window.jqCucu && jqCucu.fn.jquery;"
    jquery_version = browser.execute(script)
    if not jquery_version:
        browser.execute(load_jquery_lib())
        browser.execute("window.jqCucu = jQuery.noConflict(true);")
    text = browser.execute(
        'return jqCucu("body").children(":visible").text();'
    )
    return text


def ellipsize_filename(raw_filename):
    max_filename = 100
    new_raw_filename = normalize_filename(raw_filename)
    if len(new_raw_filename) < max_filename:
        return new_raw_filename

    ellipsis = "..."
    # save the last chars, as the ending is often important
    end_count = 40
    front_count = max_filename - (len(ellipsis) + end_count)
    ellipsized_filename = (
        new_raw_filename[:front_count]
        + ellipsis
        + new_raw_filename[-1 * end_count :]
    )
    return ellipsized_filename


def normalize_filename(raw_filename):
    normalized_filename = (
        raw_filename.replace('"', "")
        .replace("{", "")
        .replace("}", "")
        .replace("#", "")
        .replace("&", "")
    )
    return normalized_filename


def get_step_image_dir(step_index, step_name):
    """
    generate .png image file name that meets these criteria:
     - hides secrets
     - escaped
     - filename does not exceed 255 chars (OS limitation)
     - uniqueness comes from step number
    """
    escaped_step_name = CONFIG.hide_secrets(step_name).replace("/", "_")
    unabridged_dirname = f"{step_index:0>4} - {escaped_step_name}"
    dirname = ellipsize_filename(unabridged_dirname)

    return dirname


def take_saw_element_screenshot(ctx, thing, name, index, element=None):
    observed = "saw" if element else "did not see"
    prefix = "" if index == 0 else f"{humanize.ordinal(index)} "

    take_screenshot(
        ctx,
        ctx.current_step.name,
        label=f'{observed} {prefix}{thing} "{name}"',
        element=element,
    )


def take_screenshot(ctx, step_name, label="", element=None):
    step = ctx.current_step
    if not hasattr(step, "screenshots"):
        step.screenshots = []

    screenshot_dir = os.path.join(
        ctx.scenario_dir, get_step_image_dir(ctx.step_index, step_name)
    )
    if not os.path.exists(screenshot_dir):
        os.mkdir(screenshot_dir)

    if len(label) > 0:
        label = f" - {CONFIG.hide_secrets(label).replace('/', '_')}"
    filename = f"{CONFIG['__STEP_SCREENSHOT_COUNT']:0>4}{label}.png"
    filename = ellipsize_filename(filename)
    filepath = os.path.join(screenshot_dir, filename)

    if CONFIG["CUCU_SKIP_HIGHLIGHT_BORDER"] or not element:
        ctx.browser.screenshot(filepath)
    else:
        location = element.location
        border_width = 4
        x, y = location["x"] - border_width, location["y"] - border_width
        size = element.size
        width, height = size["width"], size["height"]

        position_css = f"position: absolute; top: {y}px; left: {x}px; width: {width}px; height: {height}px; z-index: 9001;"
        visual_css = "border-radius: 4px; border: 4px solid #ff00ff1c; background: #ff00ff05; filter: drop-shadow(magenta 0 0 10px);"

        script = f"""
            (function() {{ // double curly-brace to escape python f-string
                var body = document.querySelector('body');
                var cucu_border = document.createElement('div');
                cucu_border.setAttribute('id', 'cucu_border');
                cucu_border.setAttribute('style', '{position_css} {visual_css}');
                body.append(cucu_border);
            }})();
        """
        ctx.browser.execute(script)

        ctx.browser.screenshot(filepath)

        clear_highlight = """
            (function() {
                var body = document.querySelector('body');
                var cucu_border = document.getElementById('cucu_border');
                body.removeChild(cucu_border);
            })();
        """
        ctx.browser.execute(clear_highlight, element)

    screenshot = {
        "step_name": step_name,
        "label": label,
        "element": element,
        "location": f"({element.location['x']},{element.location['y']})"
        if element
        else "",
        "size": f"({element.size['width']},{element.size['height']})"
        if element
        else "",
        "filepath": filepath,
    }
    step.screenshots.append(screenshot)

    if CONFIG["CUCU_MONITOR_PNG"]:
        shutil.copyfile(filepath, CONFIG["CUCU_MONITOR_PNG"])

    CONFIG["__STEP_SCREENSHOT_COUNT"] += 1


def find_n_click_input_parent_label(ctx, input_element):
    """
    Clicks the nearest parent <label> of an input elemnt (if input is visually hidden or size is zero).
    """
    try:
        # Find the closest ancestor <label> element
        label = input_element.find_element(By.XPATH, "ancestor::label[1]")

        if label and label.is_displayed():
            ctx.browser.click(label)
            logger.debug("Successfully clicked the parent label.")
        else:
            logger.warning("Parent label is not displayed or not found.")

    except Exception as e:
        logger.error(
            f"Click on parent label failed (possibly missing label ancestor): {e}"
        )


def is_element_size_zero(element):
    size = element.size
    return size["width"] == 0 and size["height"] == 0


class TeeStream:
    """
    A stream that writes to both a file stream and captures content in an internal buffer.
    Provides file-like accessors to read the captured content.
    """

    def __init__(self, file_stream):
        self.file_stream = file_stream
        self.string_buffer = []

    def write(self, data):
        self.file_stream.write(data)
        self.string_buffer.append(data)

    def flush(self):
        self.file_stream.flush()

    def getvalue(self):
        return "".join(self.string_buffer)

    def read(self):
        return self.getvalue()

    def clear(self):
        """Clear the internal buffer."""
        self.string_buffer = []



import glob
import os
import shutil
import sys
import traceback
import urllib
from datetime import datetime
from pathlib import Path
from xml.sax.saxutils import escape as escape_

import jinja2

from cucu import format_gherkin_table, logger
from cucu.ansi_parser import parse_log_to_html
from cucu.config import CONFIG
from cucu.db import close_html_report_db, init_html_report_db
from cucu.db import feature as FeatureModel
from cucu.db import scenario as ScenarioModel
from cucu.db import step as StepModel
from cucu.utils import ellipsize_filename, get_step_image_dir


def escape(data):
    if data is None:
        return None

    return escape_(data, {'"': "&quot;"}).rstrip()


def process_tags(element):
    """
    process tags in the element provided (scenario or feature) and basically
    convert the tags to a simple @xxx representation.
    """
    prepared_tags = []

    if "tags" not in element:
        return

    for tag in element["tags"]:
        tag = f"@{tag}"

        # process custom tag handlers
        tag_handlers = CONFIG["__CUCU_HTML_REPORT_TAG_HANDLERS"].items()
        for regex, handler in tag_handlers:
            if regex.match(tag):
                tag = handler(tag)

        prepared_tags.append(tag)

    element["tags"] = " ".join(prepared_tags)


# function to left pad duration with '0' for better alphabetical sorting in html reports.
def left_pad_zeroes(elapsed_time):
    int_decimal = str(round(elapsed_time, 3)).split(".")
    int_decimal[0] = int_decimal[0].zfill(3)
    padded_duration = ".".join(int_decimal)
    return padded_duration


def generate(results, basepath, only_failures=False):
    """
    generate an HTML report for the results provided.
    """

    features = []

    db_path = os.path.join(results, "run.db")
    try:
        init_html_report_db(db_path)
        features = []

        db_features = FeatureModel.select().order_by(FeatureModel.start_at)
        logger.info(
            f"Starting to process {len(db_features)} features for report"
        )

        for db_feature in db_features:
            feature_dict = {
                "name": db_feature.name,
                "filename": db_feature.filename,
                "description": db_feature.description,
                "tags": db_feature.tags if db_feature.tags else [],
                "status": db_feature.status,
                "elements": [],
            }

            db_scenarios = (
                ScenarioModel.select()
                .where(
                    ScenarioModel.feature_run_id == db_feature.feature_run_id
                )
                .order_by(ScenarioModel.seq)
            )

            feature_has_failures = False

            if len(db_scenarios) == 0:
                logger.debug(f"Feature {db_feature.name} has no scenarios")
                continue

            for db_scenario in db_scenarios:
                scenario_dict = {
                    "name": db_scenario.name,
                    "line": db_scenario.line_number,
                    "tags": db_scenario.tags if db_scenario.tags else [],
                    "status": db_scenario.status or "passed",
                    "steps": [],
                }

                if db_scenario.status == "failed":
                    feature_has_failures = True

                db_steps = (
                    StepModel.select()
                    .where(
                        StepModel.scenario_run_id
                        == db_scenario.scenario_run_id
                    )
                    .order_by(StepModel.seq)
                )

                for db_step in db_steps:
                    step_dict = {
                        "keyword": db_step.keyword,
                        "name": db_step.name,
                        "result": {
                            "status": db_step.status or "passed",
                            "duration": db_step.duration or 0,
                            "timestamp": db_step.end_at or "",
                        },
                        "substep": db_step.is_substep,
                    }

                    if db_step.text:
                        step_dict["text"] = db_step.text

                    if db_step.table_data:
                        step_dict["table"] = db_step.table_data

                    step_dict["result"]["error_message"] = (
                        db_step.error_message.splitlines()
                        if db_step.error_message
                        else []
                    )
                    step_dict["result"]["exception"] = db_step.exception
                    step_dict["result"]["stdout"] = db_step.stdout
                    step_dict["result"]["stderr"] = db_step.stderr
                    step_dict["result"]["browser_logs"] = (
                        db_step.browser_logs.splitlines()
                    )
                    step_dict["result"]["debug_output"] = (
                        db_step.debug_output.splitlines()
                    )

                    scenario_dict["steps"].append(step_dict)

                feature_dict["elements"].append(scenario_dict)

            if feature_has_failures:
                feature_dict["status"] = "failed"
            elif only_failures and not feature_has_failures:
                continue

            features.append(feature_dict)

    finally:
        close_html_report_db()

    cucu_dir = os.path.dirname(sys.modules["cucu"].__file__)
    external_dir = os.path.join(cucu_dir, "reporter", "external")
    shutil.copytree(external_dir, os.path.join(basepath, "external"))
    shutil.copyfile(
        os.path.join(cucu_dir, "reporter", "favicon.png"),
        os.path.join(basepath, "favicon.png"),
    )

    #
    # augment existing test run data with:
    #  * features & scenarios with `duration` attribute computed by adding all
    #    step durations.
    #  * add `image` attribute to a step if it has an underlying .png image.
    #
    CONFIG.snapshot()
    reported_features = []
    for feature in features:
        feature["folder_name"] = ellipsize_filename(feature["name"])
        scenarios = []

        if feature["status"] != "untested" and "elements" in feature:
            scenarios = feature["elements"]

        if only_failures and feature["status"] != "failed":
            continue

        feature_duration = 0
        total_scenarios = 0
        total_scenarios_passed = 0
        total_scenarios_failed = 0
        total_scenarios_skipped = 0
        total_scenarios_errored = 0
        feature_started_at = None

        reported_features.append(feature)
        process_tags(feature)

        if feature["status"] not in ["skipped", "untested"]:
            # copy each feature directories contents over to the report directory
            src_feature_filepath = os.path.join(
                results, feature["folder_name"]
            )
            dst_feature_filepath = os.path.join(
                basepath, feature["folder_name"]
            )
            shutil.copytree(
                src_feature_filepath, dst_feature_filepath, dirs_exist_ok=True
            )

        for scenario in scenarios:
            CONFIG.restore()

            scenario["folder_name"] = ellipsize_filename(scenario["name"])
            scenario_filepath = os.path.join(
                basepath,
                feature["folder_name"],
                scenario["folder_name"],
            )

            scenario_configpath = os.path.join(
                scenario_filepath, "logs", "cucu.config.yaml.txt"
            )
            if os.path.exists(scenario_configpath):
                try:
                    CONFIG.load(scenario_configpath)
                except Exception as e:
                    logger.warning(
                        f"Could not reload config: {scenario_configpath}: {e}"
                    )
            else:
                logger.info(f"No config to reload: {scenario_configpath}")

            process_tags(scenario)

            sub_headers = []
            for handler in CONFIG[
                "__CUCU_HTML_REPORT_SCENARIO_SUBHEADER_HANDLER"
            ]:
                try:
                    sub_header = handler(scenario, feature)
                    if sub_header:
                        sub_headers.append(sub_header)
                except Exception:
                    logger.warning(
                        f'Exception while trying to run sub_headers hook for scenario: "{scenario["name"]}"\n{traceback.format_exc()}'
                    )
            scenario["sub_headers"] = "<br/>".join(sub_headers)

            scenario_duration = 0
            total_scenarios += 1
            total_steps = 0

            if "status" not in scenario:
                total_scenarios_skipped += 1
            elif scenario["status"] == "passed":
                total_scenarios_passed += 1
            elif scenario["status"] == "failed":
                total_scenarios_failed += 1
            elif scenario["status"] == "skipped":
                total_scenarios_skipped += 1
            elif scenario["status"] == "errored":
                total_scenarios_errored += 1

            step_index = 0
            scenario_started_at = None
            for step in scenario["steps"]:
                total_steps += 1
                image_dir = get_step_image_dir(step_index, step["name"])
                image_dirpath = os.path.join(scenario_filepath, image_dir)

                # Handle section headings with different levels (# to ####)
                if step["name"].startswith("#"):
                    # Map the count to the appropriate HTML heading (h2-h5)
                    # We use h2-h5 instead of h1-h4 so h1 can be reserved for scenario/feature titles
                    step["heading_level"] = (
                        f"h{step['name'][:4].count('#') + 1}"
                    )

                if os.path.exists(image_dirpath):
                    _, _, image_names = next(os.walk(image_dirpath))
                    images = []
                    for image_name in image_names:
                        words = image_name.split("-", 1)
                        index = words[0].strip()
                        try:
                            # Images with label should have a name in the form:
                            # 0000 - This is the image label.png
                            label, _ = os.path.splitext(words[1].strip())
                        except IndexError:
                            # Images with no label should instead look like:
                            # 0000.png
                            # so we default to the step name in this case.
                            label = step["name"]

                        images.append(
                            {
                                "src": urllib.parse.quote(
                                    os.path.join(image_dir, image_name)
                                ),
                                "index": index,
                                "label": label,
                            }
                        )
                    step["images"] = sorted(images, key=lambda x: x["index"])

                if "result" in step:
                    if step["result"]["status"] in ["failed", "passed"]:
                        if step["result"]["timestamp"]:
                            timestamp = datetime.fromisoformat(
                                step["result"]["timestamp"]
                            )
                            step["result"]["timestamp"] = timestamp

                            if scenario_started_at is None:
                                scenario_started_at = timestamp
                                scenario["started_at"] = timestamp
                            time_offset = datetime.utcfromtimestamp(
                                (
                                    timestamp - scenario_started_at
                                ).total_seconds()
                            )
                            step["result"]["time_offset"] = time_offset
                        else:
                            step["result"]["timestamp"] = ""
                            step["result"]["time_offset"] = ""

                    scenario_duration += step["result"]["duration"]

                    if "error_message" in step["result"] and step["result"][
                        "error_message"
                    ] == [None]:
                        step["result"]["error_message"] = [""]

                if "text" in step and not isinstance(step["text"], list):
                    step["text"] = [step["text"]]

                # prepare by joining into one big chunk here since we can't do it in the Jinja template
                if "text" in step:
                    text_indent = "       "
                    step["text"] = "\n".join(
                        [text_indent + '"""']
                        + [f"{text_indent}{x}" for x in step["text"]]
                        + [text_indent + '"""']
                    )

                # prepare by joining into one big chunk here since we can't do it in the Jinja template
                if "table" in step:
                    step["table"] = format_gherkin_table(
                        step["table"]["rows"],
                        step["table"]["headings"],
                        "       ",
                    )

                step_index += 1
            logs_dir = os.path.join(scenario_filepath, "logs")

            if os.path.exists(logs_dir):
                log_files = []
                for log_file in glob.iglob(os.path.join(logs_dir, "*.*")):
                    log_filepath = log_file.removeprefix(
                        f"{scenario_filepath}/"
                    )

                    if ".console." in log_filepath and scenario_started_at:
                        log_filepath += ".html"

                    log_files.append(
                        {
                            "filepath": log_filepath,
                            "name": os.path.basename(log_file),
                        }
                    )

                scenario["logs"] = log_files

            scenario["total_steps"] = total_steps
            if scenario_started_at is None:
                scenario["started_at"] = ""
            else:
                if feature_started_at is None:
                    feature_started_at = scenario_started_at
                    feature["started_at"] = feature_started_at

                scenario["time_offset"] = datetime.utcfromtimestamp(
                    (scenario_started_at - feature_started_at).total_seconds()
                )

                for log_file in [
                    x for x in log_files if ".console." in x["name"]
                ]:
                    log_file_filepath = os.path.join(
                        scenario_filepath, "logs", log_file["name"]
                    )

                    input_file = Path(log_file_filepath)
                    output_file = Path(log_file_filepath + ".html")
                    output_file.write_text(
                        parse_log_to_html(
                            input_file.read_text(encoding="utf-8")
                        ),
                        encoding="utf-8",
                    )

            scenario["duration"] = left_pad_zeroes(scenario_duration)
            feature_duration += scenario_duration

        if feature_started_at is None:
            feature["started_at"] = ""

        feature["total_steps"] = sum([x["total_steps"] for x in scenarios])
        feature["duration"] = left_pad_zeroes(
            sum([float(x["duration"]) for x in scenarios])
        )

        feature["total_scenarios"] = total_scenarios
        feature["total_scenarios_passed"] = total_scenarios_passed
        feature["total_scenarios_failed"] = total_scenarios_failed
        feature["total_scenarios_skipped"] = total_scenarios_skipped
        feature["total_scenarios_errored"] = total_scenarios_errored

    keys = [
        "total_scenarios",
        "total_scenarios_passed",
        "total_scenarios_failed",
        "total_scenarios_skipped",
        "total_scenarios_errored",
        "duration",
    ]
    grand_totals = {}
    for k in keys:
        grand_totals[k] = sum([float(x[k]) for x in reported_features])

    package_loader = jinja2.PackageLoader("cucu.reporter", "templates")
    templates = jinja2.Environment(loader=package_loader)  # nosec

    def urlencode(string):
        """
        handles encoding specific characters in the names of features/scenarios
        so they can be used in a URL. NOTICE: we're not handling spaces since
        the browser handles those already.

        """
        return (
            string.replace('"', "%22").replace("'", "%27").replace("#", "%23")
        )

    templates.globals.update(escape=escape, urlencode=urlencode)

    index_template = templates.get_template("index.html")
    rendered_index_html = index_template.render(
        features=reported_features,
        grand_totals=grand_totals,
        title="Cucu HTML Test Report",
        basepath=basepath,
        dir_depth="",
    )

    index_output_filepath = os.path.join(basepath, "index.html")
    with open(index_output_filepath, "wb") as output:
        output.write(rendered_index_html.encode("utf8"))

    flat_template = templates.get_template("flat.html")
    rendered_flat_html = flat_template.render(
        features=reported_features,
        grand_totals=grand_totals,
        title="Flat HTML Test Report",
        basepath=basepath,
        dir_depth="",
    )

    flat_output_filepath = os.path.join(basepath, "flat.html")
    with open(flat_output_filepath, "wb") as output:
        output.write(rendered_flat_html.encode("utf8"))

    feature_template = templates.get_template("feature.html")

    for feature in reported_features:
        feature_basepath = os.path.join(basepath, feature["folder_name"])
        os.makedirs(feature_basepath, exist_ok=True)

        scenarios = []
        if feature["status"] != "untested" and "elements" in feature:
            scenarios = feature["elements"]

        rendered_feature_html = feature_template.render(
            feature=feature,
            scenarios=scenarios,
            dir_depth="",
            title=feature.get("name", "Cucu results"),
        )

        feature_output_filepath = os.path.join(
            basepath, f"{feature['name']}.html"
        )

        with open(feature_output_filepath, "wb") as output:
            output.write(rendered_feature_html.encode("utf8"))

        scenario_template = templates.get_template("scenario.html")

        for scenario in scenarios:
            steps = scenario["steps"]
            scenario_basepath = os.path.join(
                feature_basepath, scenario["folder_name"]
            )
            os.makedirs(scenario_basepath, exist_ok=True)

            scenario_output_filepath = os.path.join(
                scenario_basepath, "index.html"
            )

            rendered_scenario_html = scenario_template.render(
                basepath=results,
                feature=feature,
                path_exists=os.path.exists,
                scenario=scenario,
                steps=steps,
                title=scenario.get("name", "Cucu results"),
                dir_depth="../../",
            )

            with open(scenario_output_filepath, "wb") as output:
                output.write(rendered_scenario_html.encode("utf8"))

    return os.path.join(basepath, "flat.html")



import logging
import os
import re

import chromedriver_autoinstaller
import geckodriver_autoinstaller
import urllib3
from selenium import webdriver
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.common.by import By
from selenium.webdriver.edge.options import Options as EdgeOptions
from selenium.webdriver.firefox.options import Options as FirefoxOptions

from cucu import config, edgedriver_autoinstaller, logger
from cucu.browser.core import Browser
from cucu.browser.frames import search_in_all_frames


class DisableLogger:
    def __enter__(self):
        logging.disable(logging.CRITICAL)

    def __exit__(self, exit_type, exit_value, exit_traceback):
        logging.disable(logging.NOTSET)


def init():
    """
    initialize any selenium specific needs
    """
    if config.CONFIG["CUCU_BROWSER"] == "chrome":
        try:
            with DisableLogger():
                # auto install chromedriver if not present
                chromedriver_autoinstaller.install()
        except:  # noqa: E722
            logging.debug("unable to auto install chromedriver")

    if config.CONFIG["CUCU_BROWSER"] == "firefox":
        # https://github.com/mozilla/geckodriver/issues/330
        logger.warning("browser console logs not available on firefox")
        geckodriver_autoinstaller.install()

    if config.CONFIG["CUCU_BROWSER"] == "edge":
        edgedriver_autoinstaller.install()


class Selenium(Browser):
    def __init__(self):
        self.driver = None

    def open(
        self, browser, headless=False, selenium_remote_url=None, detach=False
    ):
        if selenium_remote_url is None:
            init()

        timeout = float(config.CONFIG["CUCU_SELENIUM_DEFAULT_TIMEOUT_S"])
        os.environ["GLOBAL_DEFAULT_TIMEOUT"] = str(timeout)

        height = config.CONFIG["CUCU_BROWSER_WINDOW_HEIGHT"]
        width = config.CONFIG["CUCU_BROWSER_WINDOW_WIDTH"]
        cucu_downloads_dir = config.CONFIG["CUCU_BROWSER_DOWNLOADS_DIR"]
        ignore_ssl_certificate_errors = config.CONFIG[
            "CUCU_IGNORE_SSL_CERTIFICATE_ERRORS"
        ]

        if browser.startswith("chrome"):
            options = ChromeOptions()
            options.add_experimental_option("detach", detach)

            prefs = {"download.default_directory": cucu_downloads_dir}
            options.add_experimental_option("prefs", prefs)

            options.add_argument(f"--window-size={width},{height}")
            options.add_argument("--disable-dev-shm-usage")

            if headless:
                options.add_argument("--headless")

            if ignore_ssl_certificate_errors:
                options.add_argument("ignore-certificate-errors")
            options.set_capability("goog:loggingPrefs", {"browser": "ALL"})

            if selenium_remote_url is not None:
                logger.debug(f"webdriver.Remote init: {selenium_remote_url}")
                try:
                    self.driver = webdriver.Remote(
                        command_executor=selenium_remote_url,
                        options=options,
                    )
                except urllib3.exceptions.ReadTimeoutError:
                    print("*" * 80)
                    print(
                        "* unable to connect to the remote selenium setup,"
                        " you may need to restart it"
                    )
                    print("*" * 80)
                    print("")
                    raise
            else:
                logger.debug("webdriver.Chrome init")
                self.driver = webdriver.Chrome(
                    options=options,
                )

        elif browser.startswith("firefox"):
            options = FirefoxOptions()

            options.set_preference("browser.download.folderList", 2)
            options.set_preference(
                "browser.download.manager.showWhenStarting", False
            )
            options.set_preference("browser.download.dir", cucu_downloads_dir)
            options.set_preference(
                "browser.helperApps.neverAsk.saveToDisk",
                "images/jpeg, application/pdf, application/octet-stream, text/plain",
            )

            if ignore_ssl_certificate_errors:
                options.accept_insecure_certs = True

            options.add_argument(f"--width={width}")
            options.add_argument(f"--height={height}")

            # TODO: re-enable once Firefox supports this
            # options.set_capability("loggingPrefs", {"browser": "ALL"})

            if headless:
                options.add_argument("--headless")

            if selenium_remote_url is not None:
                logger.debug(f"webdriver.Remote init: {selenium_remote_url}")
                try:
                    self.driver = webdriver.Remote(
                        command_executor=selenium_remote_url,
                        options=options,
                    )
                except urllib3.exceptions.ReadTimeoutError:
                    print("*" * 80)
                    print(
                        "* unable to connect to the remote selenium setup,"
                        " you may need to restart it"
                    )
                    print("*" * 80)
                    print("")
                    raise
            else:
                logger.debug("webdriver.Firefox init")
                self.driver = webdriver.Firefox(
                    options=options,
                )

        elif browser.startswith("edge"):
            options = EdgeOptions()

            options.add_experimental_option(
                "prefs", {"download.default_directory": cucu_downloads_dir}
            )

            if headless:
                options.use_chromium = True
                options.add_argument("--headless")

            if ignore_ssl_certificate_errors:
                options.set_capability("acceptSslCerts", True)

            if selenium_remote_url is not None:
                logger.debug(f"webdriver.Remote init: {selenium_remote_url}")
                try:
                    self.driver = webdriver.Remote(
                        command_executor=selenium_remote_url,
                        options=options,
                    )
                except urllib3.exceptions.ReadTimeoutError:
                    print("*" * 80)
                    print(
                        "* unable to connect to the remote selenium setup,"
                        " you may need to restart it"
                    )
                    print("*" * 80)
                    print("")
                    raise
            else:
                logger.debug("webdriver.Edge init")
                edgedriver_filepath = (
                    edgedriver_autoinstaller.utils.download_edgedriver()
                )
                self.driver = webdriver.Edge(
                    executable_path=edgedriver_filepath,
                    options=options,
                )

        else:
            raise Exception(f"unknown browser {browser}")

        self.driver.set_window_size(width, height)

    def get_log(self):
        if config.CONFIG["CUCU_BROWSER"] == "firefox":
            return []

        return self.driver.get_log("browser")

    def get_current_url(self):
        return self.driver.current_url

    def navigate(self, url):
        self.driver.get(url)
        self.wait_for_page_to_load()

    def switch_to_next_tab(self):
        window_handles = self.driver.window_handles
        window_handle = self.driver.current_window_handle
        window_handle_index = window_handles.index(window_handle)

        if window_handle_index == len(window_handles) - 1:
            raise RuntimeError("no next browser tab available")
        self.driver.switch_to.window(window_handles[window_handle_index + 1])

    def switch_to_previous_tab(self):
        window_handles = self.driver.window_handles
        window_handle = self.driver.current_window_handle
        window_handle_index = window_handles.index(window_handle)

        if window_handle_index == 0:
            raise RuntimeError("no previous browser tab available")
        self.driver.switch_to.window(window_handles[window_handle_index - 1])

    def switch_to_nth_tab(self, tab_number):
        print(f"tab_number, {tab_number}")
        window_handles = self.driver.window_handles
        total_tabs = len(window_handles)
        if tab_number > total_tabs:
            raise RuntimeError(f"no {tab_number} browser tab available")
        self.driver.switch_to.window(window_handles[tab_number])

    def switch_to_tab_that_matches_regex(self, title_pattern):
        window_handles = self.driver.window_handles

        for handle in window_handles:
            self.driver.switch_to.window(handle)
            if re.search(title_pattern, self.driver.title):
                return

        raise Exception(f"No tab title matches pattern: {title_pattern}")

    def get_tab_info(self):
        window_handles = self.driver.window_handles
        current_window = self.driver.current_window_handle
        window_handle_index = window_handles.index(current_window)

        return {
            "tab_count": len(window_handles),
            "index": window_handle_index,
            "title": self.driver.title,
            "url": self.driver.current_url,
        }

    def get_all_tabs_info(self):
        window_handles = self.driver.window_handles
        current_window = self.driver.current_window_handle
        tabs_info = []

        for handle in window_handles:
            self.driver.switch_to.window(handle)
            tabs_info.append(
                {
                    "title": self.driver.title,
                    "url": self.driver.current_url,
                }
            )

        # Switch back to the original window
        self.driver.switch_to.window(current_window)

        return tabs_info

    def back(self):
        self.driver.back()
        self.wait_for_page_to_load()

    def refresh(self):
        self.driver.refresh()
        self.wait_for_page_to_load()

    def title(self):
        return self.driver.title

    def css_find_elements(self, selector):
        def find_elements_in_frame():
            elements = self.driver.find_elements(By.CSS_SELECTOR, selector)

            def visible(element):
                return element.is_displayed()

            return list(filter(visible, elements))

        return search_in_all_frames(self, find_elements_in_frame)

    def execute(self, javascript, *args):
        return self.driver.execute_script(javascript, *args)

    def click(self, element):
        element.click()
        # let cucu's own wait for page to load checks run
        self.wait_for_page_to_load()

    def switch_to_default_frame(self):
        self.driver.switch_to.default_content()

    def switch_to_frame(self, frame):
        self.driver.switch_to.frame(frame)

    def screenshot(self, filepath):
        self.driver.get_screenshot_as_file(filepath)

    def close_window(self):
        window_handles = self.driver.window_handles
        window_handle = self.driver.current_window_handle
        window_handle_index = window_handles.index(window_handle)

        self.driver.close()
        self.driver.switch_to.window(window_handles[window_handle_index - 1])

    def download_mht(self, target_filepath):
        if self.driver is None:
            logger.warning(
                "No active browser; will not attempt to download .mht file."
            )
            return

        browser_name = self.driver.name.lower()
        if "chrome" not in browser_name:
            logger.warning(
                "The web driver is not using Chrome as a web browser"
                f", but {browser_name}. This browser does not support"
                "dowloading .mht files; will not attempt to download one."
            )
            return

        mht_data = None
        if self.driver._is_remote:
            cdp_url = f"{self.driver.command_executor._client_config.remote_server_addr}/session/{self.driver.session_id}/chromium/send_command_and_get_result"
            cdp_request_body = '{"cmd": "Page.captureSnapshot", "params": {}}'
            cdp_response = self.driver.command_executor._request(
                "POST", cdp_url, cdp_request_body
            )
            # Flaky: Adding try catch block to handle the situation where we do not get a dictionary object.
            # Could not reproduce this in local. So far we have seen this issue only when running on a remote web driver.
            try:
                mht_data = cdp_response.get("value").get("data")
            except Exception as e:
                logger.error(f'object "{cdp_response.get("value")}" : {e}')
        else:
            mht_response = self.driver.execute_cdp_cmd(
                "Page.captureSnapshot", {}
            )
            mht_data = mht_response.get("data")

        if mht_data is None:
            logger.warning(
                "Something unexpected has happened: fetched MHT data, but that data was empty. Not writing MHT file."
            )
            return

        with open(target_filepath, "w") as mht_file:
            mht_file.write(mht_data)

    def quit(self):
        self.driver.quit()
